package dev.pfaff.altargui.util;

import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Signal;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.invoke.VarHandle;
import java.util.function.BiFunction;

import static java.lang.invoke.MethodHandles.filterCoordinates;
import static java.lang.invoke.MethodHandles.filterValue;

public final class WidgetStateSignalManager<T, S extends Signal> {
	private final VarHandle VH_signal;
	private final VarHandle VH_listener;
	private final BiFunction<T, S, ListenerDisposer> listenerFactory;

	public WidgetStateSignalManager(MethodHandles.Lookup lookup, String signalField, Class<S> signalType,
									String listenerField, BiFunction<T, S, ListenerDisposer> listenerFactory) {
		var receiverType = lookup.lookupClass();
		var MH_castReceiver = MethodHandles.explicitCastArguments(MethodHandles.identity(receiverType),
				MethodType.methodType(receiverType, Object.class));
		var MH_castSignalTypeToSignal = MethodHandles.explicitCastArguments(MethodHandles.identity(Signal.class),
				MethodType.methodType(Signal.class, signalType));
		var MH_castSignalToSignalType = MethodHandles.explicitCastArguments(MethodHandles.identity(signalType),
				MethodType.methodType(signalType, Signal.class));
		VarHandle VH_signal, VH_listener;
		try {
			VH_signal = lookup.findVarHandle(receiverType, signalField, signalType);
			VH_listener = lookup.findVarHandle(receiverType, listenerField, ListenerDisposer.class);
		} catch (NoSuchFieldException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
		this.VH_signal = filterValue(filterCoordinates(VH_signal.withInvokeExactBehavior(), 0, MH_castReceiver),
				MH_castSignalToSignalType,
				MH_castSignalTypeToSignal);
		this.VH_listener = filterCoordinates(VH_listener.withInvokeExactBehavior(), 0, MH_castReceiver);
		this.listenerFactory = listenerFactory;
	}

	private void setNewListener(T receiver, S signal) {
		this.VH_listener.set(receiver, this.listenerFactory.apply(receiver, signal));
	}

	public void configure(T receiver, S newSignal) {
		if ((S) this.VH_signal.get(receiver) != newSignal) {
			var currentListener = (ListenerDisposer) this.VH_listener.get(receiver);
			if (currentListener != null) currentListener.close();
			this.VH_signal.set(receiver, newSignal);
			setNewListener(receiver, newSignal);
		}
	}

	public void onMount(T receiver) {
		if ((ListenerDisposer) this.VH_listener.get(receiver) == null) {
			var signal = (S) this.VH_signal.get(receiver);
			if (signal != null) {
				setNewListener(receiver, signal);
			}
		}
	}

	public void onUnmount(T receiver) {
		var listener = (ListenerDisposer) this.VH_listener.get(receiver);
		if (listener != null) {
			listener.close();
			this.VH_listener.set(receiver, (ListenerDisposer) null);
			this.VH_signal.set(receiver, (Signal) null);
		}
	}
}
