package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.LayoutContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.function.CharConsumer;
import dev.pfaff.altargui.function.CharProvider;
import dev.pfaff.altargui.pathfinder.fontkit.FKFontRef;
import dev.pfaff.altargui.pathfinder.fontkit.FKGlyphLoadableTexture;

import java.util.Objects;

import static dev.pfaff.altargui.Key.intKey;
import static dev.pfaff.altargui.widgets.MultiChildWidget.row;

record FKTextWidgetImpl(CharProvider charProvider, FKFontRef font, float size) implements FKTextWidget, MultiChildWidget.Builder {
	FKTextWidgetImpl {
		Objects.requireNonNull(charProvider);
		Objects.requireNonNull(font);
	}

	@Override
	public Widget build() {
		return row(this);
	}

	@Override
	public void build(LayoutContext context) {
		this.charProvider.provide(new CharConsumerImpl(this.font, this.size, context));
	}

	private static final class CharConsumerImpl implements CharConsumer {
		private final FKFontRef font;
		private final float size;
		private final LayoutContext context;
		private int i;

		public CharConsumerImpl(FKFontRef font, float size, LayoutContext context) {
			this.font = font;
			this.size = size;
			this.context = context;
		}

		@Override
		public void accept(char c) {
			this.context.layout(intKey(this.i++),
					RectWidget.textured(new FKGlyphLoadableTexture(this.font, c)).withAbsoluteSize(this.size));
		}
	}
}
